@@ -1,6 +1,6 @@
 # perldoc RRSIG.pm for documentation.
 # Specs: RFC 2535 section 4
-# $Id: RRSIG.pm,v 1.1 2003/08/27 14:09:25 olaf Exp $
+# $Id: RRSIG.pm 527 2005-12-09 10:51:06Z olaf $
 
 package Net::DNS::RR::RRSIG;
 
@@ -8,6 +8,7 @@ use vars qw(@ISA $VERSION @EXPORT );
 
 use Net::DNS;
 use Carp;
+use bytes;
 
 use Crypt::OpenSSL::DSA;
 use Crypt::OpenSSL::RSA;
@@ -18,36 +19,38 @@ use File::Basename;
 use MIME::Base64;
 use Math::BigInt;
 use Time::Local;
-use Digest::SHA1 qw (sha1);
-
+use Digest::SHA qw (sha1);
 
 
 
 
 
 #
-# Most of the cryptovariables should be interpred as unsigne
+# Most of the cryptovariables should be interpred as unsigned
 #
 #
 
 
 require Exporter;
 
-$VERSION = do { my @r=(q$Revision: 1.1 $=~/\d+/g); sprintf "%d."."%03d"x$#r,@r };
+$VERSION = do { my @r=(q$Revision: 527 $=~/\d+/g); sprintf "%d."."%03d"x$#r,@r };
 @ISA = qw (
 	   Exporter
-	 Net::DNS::RR
+  	 Net::DNS::RR
+  	 Net::DNS::SEC
 	   );
 
-@EXPORT = qw (
+
+@EXPORT = qw (          
 	      );
 
 
-use strict;
-my $debug=0;
-my $crypt_open_ssl=1;
 
 
+use strict;
+my $crypt_open_ssl=1;
+my $debug=0;
+
 
 
 
@@ -74,14 +77,14 @@ sub new {
 					   $expt[3] ,$expt[2] , $expt[1]  , 
 					   $expt[0]);
 	my @inct=gmtime(unpack("N",substr($$data,$offsettosiginc,4)));
-	$self->{"siginceptation"}=  sprintf ("%d%02d%02d%02d%02d%02d",
+	$self->{"siginception"}=  sprintf ("%d%02d%02d%02d%02d%02d",
 					     $inct[5]+1900 ,$inct[4]+1 , 
 					     $inct[3] ,$inct[2] , $inct[1]  ,
 					     $inct[0]);
 	$self->{"keytag"}=unpack("n",substr($$data,$offsettokeytag,2));
 	my($signame,$sigoffset) = Net::DNS::Packet::dn_expand
 	    ($data, $offsettosignm);
-	$self->{"signame"}=lc($signame);
+	$self->{"signame"}=lc($signame).".";  #Add a trailing dot..dn_expand does not do that.
 	my($sigmaterial)=substr($$data,$sigoffset,
 				($self->{"rdlength"}-$sigoffset+$offset));
 	$self->{"sigbin"}=$sigmaterial;
@@ -103,19 +106,19 @@ sub new_from_string {
 	$string =~ s/\n//mg;
 	my ($typecovered, $algoritm,
 	    $labels, $orgttl, $sigexpiration,
-	    $siginceptation, $keytag,$signame,$sig) = 
+	    $siginception, $keytag,$signame,$sig) = 
 		$string =~ 
 		    /^\s*(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\S+)\s+(.*)/;
 	croak (" Invallid RRSIG RR, check your fomat ") if !$keytag;
 	$sig =~ s/\s*//g;
 	$self->{"typecovered"}= $typecovered;
-	$self->{"algorithm"}= $algoritm;
+	$self->{"algorithm"}= Net::DNS::SEC->algorithm($algoritm);
 	$self->{"labels"}= lc($labels);
 	$self->{"orgttl"}= $orgttl;
 	_checktimeformat($sigexpiration);
-	_checktimeformat($siginceptation);
+	_checktimeformat($siginception);
 	$self->{"sigexpiration"}=  $sigexpiration;
-	$self->{"siginceptation"}= $siginceptation;
+	$self->{"siginception"}= $siginception;
 	$self->{"keytag"}= $keytag;
 	$self->{"signame"}= lc($signame);
 	$self->{"sig"}= $sig;
@@ -131,11 +134,11 @@ sub rdatastr {
 	my $rdatastr;
 	if (exists $self->{"typecovered"}) {
 	    $rdatastr  = $self->{typecovered};
-	    $rdatastr .= "  "  . "$self->{algorithm}";
+	    $rdatastr .= "  "  . $self->algorithm;
 	    $rdatastr .= "  "  . "$self->{labels}";
 	    $rdatastr .= "  "  . "$self->{orgttl}";
 	    $rdatastr .= "  "  . "$self->{sigexpiration}";
-	    $rdatastr .= " (\n\t\t\t"  . "$self->{siginceptation}";
+	    $rdatastr .= " (\n\t\t\t"  . "$self->{siginception}";
 	    $rdatastr .= " "  . "$self->{keytag}";
 	    $rdatastr .= "  "  . "$self->{signame}";
 	    # do some nice formatting
@@ -159,18 +162,19 @@ sub rr_rdata_without_sigbin {
 
     if (exists $self->{"typecovered"}) {
 	$rdata  = pack("n",_string2type($self->{typecovered}));
-	$rdata .= pack("C",$self->{algorithm});
+	$rdata .= pack("C",$self->algorithm);
 	$rdata .= pack("C",$self->{"labels"});
 	$rdata .= pack("N",$self->{"orgttl"});
+
 	$self->{"sigexpiration"} =~ /(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})/;
 	$rdata .= pack("N",timegm ($6, $5, $4, $3, $2-1, $1-1900));
 
-	$self->{"siginceptation"} =~ /(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})/;
+	$self->{"siginception"} =~ /(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})/;
 	$rdata .= pack("N",timegm ($6, $5, $4, $3, $2-1, $1-1900));
 	$rdata .= pack("n",$self->{"keytag"});
 	# Since we will need canonical and expanded names while checking 
 	# we do not use the packet->dn_comp here but use RFC1035 p10.
-	{   my @dname= split /\./,lc($self->{"signame"});
+	{   my @dname= split /\./,lc($self->{"signame"});  #/ emacs fontlock
 	    for (my $i=0;$i<@dname;$i++){
 		$rdata .= pack ("C",length $dname[$i] );
 		$rdata .= $dname[$i] ;
@@ -201,7 +205,8 @@ sub rr_rdata {
 
 sub create {
     my ($class,  $datarrset, $priv_key, %args) = @_;
-    
+
+
     # This method returns a sigrr with the signature over the
     # datatrrset (an array of RRs) made with the private key stored in
     # the $key_file.
@@ -217,9 +222,12 @@ sub create {
 	$Private=Net::DNS::SEC::Private->new($priv_key);
     }
 
-    die "Create did not manage obtain a Net::DNS::SEC::Private object "unless (UNIVERSAL::isa($Private,"Net::DNS::SEC::Private"));
-
-    $self->{"algorithm"}=$Private->algorithm;
+    unless (UNIVERSAL::isa($Private,"Net::DNS::SEC::Private")){
+	$self->{"sigerrstr"}= "Create did not manage to parse a private key into a Net::DNS::SEC::Private object ";
+	return (0);
+	    
+    }
+    $self->{"algorithm"}=Net::DNS::SEC->algorithm($Private->algorithm);
     $self->{"keytag"}=$Private->keytag;
     $self->{"signame"}=$Private->signame;
 
@@ -238,40 +246,40 @@ sub create {
     $self->{"class"}="IN";
 
 
-    if ($args{ttl}){
+    if (defined ($args{ttl})){
 	print "\nSetting TTL to ".  $args{"ttl"} if $debug;
 	$self->{"ttl"}= $args{"ttl"};
     }else{
-	$self->{"ttl"}= 3600;
+	$self->{"ttl"}= $datarrset->[0]->ttl;
     }
 
     $self->{"typecovered"}=$datarrset->[0]->type;  #Sanity checks elsewhere
 
 
 
-    if ($args{response}){
+    if (defined ($args{response})){
 	$self->{"response"}=$args{"response"};
     }
 
-    if ($args{"sigin"}){
+    if (defined($args{"sigin"})){
 	_checktimeformat($args{"sigin"});
-	print "\nSetting siginceptation to " . $args{"sigin"} if $debug;
-	$self->{"siginceptation"} =$args{"sigin"};
+	print "\nSetting siginception to " . $args{"sigin"} if $debug;
+	$self->{"siginception"} =$args{"sigin"};
     }else{
 	my @inct=gmtime(time);
 	my $currentdatestring=  sprintf ("%d%02d%02d%02d%02d%02d",
 					 $inct[5]+1900 ,$inct[4]+1 , 
 					 $inct[3] ,$inct[2] , $inct[1]  ,
 					 $inct[0]);	
-	$self->{"siginceptation"} = $currentdatestring ;
+	$self->{"siginception"} = $currentdatestring ;
     }
 
     # This will fail if the dateformat is not correct...
-    $self->{"siginceptation"} =~ 
+    $self->{"siginception"} =~ 
 	/(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})/ ;
     my $siginc_time=timegm ($6, $5, $4, $3, $2-1, $1-1900);
 
-    if ($args{"sigval"}){ #sigexpiration set by siginception + sigval
+    if (defined($args{"sigval"})){ #sigexpiration set by siginception + sigval
 	my @inct;
 
 
@@ -284,13 +292,15 @@ sub create {
 					   $inct[0]);	
     }elsif ($args{"sigex"}) { #sigexpiration set by the argument
 	_checktimeformat($args{"sigex"});
-	if ( $self->{"siginceptation"} > $args{"sigex"} ){
+
+	if ( $self->{"siginception"} > $args{"sigex"} ){
 	    croak "Signature can only expire after it has been incepted (".
-		$args{"sigex"} . "<" . $self->{"siginceptation"} .
+		$args{"sigex"} . "<" . $self->{"siginception"} .
 		    ")";
 	}
-	print "\nSetting sigexpiration to " . $args{"sigexp"} if $debug;
+	print "\nSetting sigexpiration to " . $args{"sigex"} if $debug;
 	$self->{"sigexpiration"}=$args{"sigex"} ;
+
     }else{ 
 	my @inct;
 
@@ -305,10 +315,11 @@ sub create {
 
 
     my  $labels=$datarrset->[0]->name;
-    $labels =~ s/\.^//;  # remove trailing dot.
-    my @labels= split /\./ , $labels;
+    $labels =~ s/\.$//;  # remove trailing dot.
+    $labels =~ s/^\*\.//;  # remove initial asterisk label
+    my @labels= split /\./ , $labels;    # / emacs font-lock-mode	
     $self->{"labels"}= scalar(@labels);
-	
+
 
     # All the TTLs need to be the same in the data RRset.
     if ( @{$datarrset}>1 ){
@@ -318,8 +329,7 @@ sub create {
 	    }
 	}
     }
-  
-    $self->{"orgttl"}=0;
+    $self->{"orgttl"}=$datarrset->[0]->{"ttl"};
 
     $self->{"sig"}=  "NOTYETCALCULATED";  # This is what we'll do in a bit...
     $self->{"sigbin"}= decode_base64($self->{"sig"});
@@ -329,7 +339,7 @@ sub create {
     # more ways to do things)
 
     bless $self, $class;
-    
+
     my $sigdata=$self->_CreateSigData($datarrset);
 
 
@@ -339,8 +349,8 @@ sub create {
     
     #
     # Enjoy the crypto
-    if ($self->{"algorithm"} == 1 || $self->{"algorithm"} == 5) {  #RSA
-	if (! ($Private->algorithm == 1 || $self->algorithm == 5 )) {
+    if ($self->algorithm == 1 || $self->algorithm == 5) {  #RSA
+	if (! ($Private->algorithm == 1 || $Private->algorithm == 5 )) {
 	    die "Private key mismatch, not RSAMD5 or RSASHA.";
 	    
 	}
@@ -349,7 +359,7 @@ sub create {
 	$self->{"private_key"}=$Private->privatekey;
 	eval {
 	    $rsa_priv->use_pkcs1_oaep_padding;
-	    if ($self->{"algorithm"} == 1) {
+	    if ($self->algorithm == 1) {
 		$rsa_priv->use_md5_hash;
 	    } else {
 		$rsa_priv->use_sha1_hash;
@@ -364,7 +374,7 @@ sub create {
 
 	print "\n SIGNED" if $debug ;
 	
-    }elsif ($self->{"algorithm"} == 3){  #DSA
+    }elsif ($self->algorithm == 3){  #DSA
 	$self->{"private_key"}=$Private->privatekey;
 	my $private_dsa=$Private->privatekey;
 
@@ -410,7 +420,7 @@ sub create {
 
 
 sub verify {
-    my ($self, $dataref, $keyrr) = @_;
+    my ($self, $dataref, $keyrrref ) = @_;
 
     # Reminder...
 
@@ -419,13 +429,68 @@ sub verify {
     # object in the array...  @{$dataref} is length of the array when
     # called in numerical context
 
+    # $keyref is eiter a reference to an array of keys or a a key object.
+
     # if $dataref is not a reference it contains a string with data to be 
     # verified using SIG0
     
     my $sigzero_verify=0;
     my $packet_verify=0;
     my $rrarray_verify=0;
+
+    my $keyrr; # This will be used to store the key to which we want to 
+               # verify.
    
+
+    print "Second argument is of class".ref($keyrrref)."\n" if $debug;;
+    if (ref($keyrrref) eq "ARRAY"){
+	#  We will recurse for each key that matches algorithm and key-id 
+	#  we return when there is a succesful verification.
+        #  If not we'll continue so that we even survive key-id collission.
+	#  The downside of this is that the error string only matches the
+	#  last error.
+	my @keyarray=@{$keyrrref};
+	my $errorstring="";
+	my $i=0;
+	print "Itterating over " . @keyarray ." keys \n" if $debug;
+      KEYRR: foreach my $keyrr (@keyarray) {
+	  $i++;
+	  unless ($keyrr->algorithm == $self->algorithm){
+	      print "key $i: algorithm does not match\n" if $debug;
+	      $errorstring.="key $i: algorithm does not match ";
+	      next KEYRR;
+	  }
+	  unless ($keyrr->keytag == $self->keytag){
+	      print "key $i: keytag does not match (".$keyrr->keytag." ".$self->keytag.")\n" if $debug;
+	      $errorstring.="key $i: keytag does not match ";
+	      next KEYRR ;
+	      
+	  }
+	  my $result=$self->verify($dataref,$keyrr);
+	  print "key $i:".$self->vrfyerrstr if $debug;
+	  $errorstring.="key $i:".$self->vrfyerrstr." ";
+	  next KEYRR unless $result;
+	  $self->{"vrfyerrstr"}="No Error";
+	  return $result;
+      }
+	$self->{"vrfyerrstr"}=$errorstring;	  
+	return (0);
+    }elsif(ref($keyrrref) eq 'Net::DNS::RR::DNSKEY' ||
+	   ref($keyrrref) eq 'Net::DNS::RR::KEY' # we are liberal...
+	){
+	# substitute and continue processing after this conditional
+	$keyrr=$keyrrref;
+	print "Validating using key with keytag:".$keyrr->keytag." \n" if $debug;
+
+    }else{
+	# Error condition
+	$self->{"vrfyerrstr"} = "You are trying to pass ".ref($keyrrref) ." data for a key";
+	return (0);
+    }
+
+
+
+
     print "Verifying data of class:".  ref( $dataref) . "\n" if $debug;
     $sigzero_verify=1 unless (ref($dataref));
     if (! $sigzero_verify ){
@@ -446,6 +511,8 @@ sub verify {
 	}
     }
 
+
+
     $self->{"vrfyerrstr"}="---- Unknown Error Condition ------";
     print "\n ------------------------------- RRSIG DEBUG  -----------------\n"  if $debug;
     print "Reference: ".ref($dataref) if $debug;;
@@ -533,9 +600,9 @@ sub verify {
 					     $inct[5]+1900 ,$inct[4]+1 , 
 					     $inct[3] ,$inct[2] , $inct[1]  ,
 					     $inct[0]);	
-	if ($self->{"siginceptation"} > $currentdatestring ){
+	if ($self->{"siginception"} > $currentdatestring ){
 	    $self->{"vrfyerrstr"}= "Signature may only be used in the future; after " .
-		$self->{"siginceptation"} ;
+		$self->{"siginception"} ;
 	    return 0;
 	}elsif($self->{"sigexpiration"} < $currentdatestring ){
 	    $self->{"vrfyerrstr"}= "Signature has expired since: " .
@@ -561,8 +628,8 @@ sub verify {
 #
 sub _type2string {
     my $index=shift;
-    if( exists $Net::DNS::typesbyval{$index}){
-	return $Net::DNS::typesbyval{$index} ;
+    if( Net::DNS::typesbyval($index)){
+	return Net::DNS::typesbyval($index) ;
     }else{
 	return "UNKNOWN TYPE";
     }
@@ -570,10 +637,10 @@ sub _type2string {
 
 sub _string2type {
     my $index=shift;
-        if( exists $Net::DNS::typesbyname{uc($index)}){
-	return $Net::DNS::typesbyname{uc($index)} ;
+        if( Net::DNS::typesbyname(uc($index))){
+	return Net::DNS::typesbyname(uc($index)) ;
     }else{
-	carp "UNKNOWN QTYPE, cannot continue ";
+	carp "UNKNOWN TYPE, cannot continue ";
     }
 }
 
@@ -774,7 +841,7 @@ sub _verifyRSA {
 
 sub _CreateSigData {
     # this is the data that will be  signed, it will be fed to the
-    # verifier. See RFC2535 4.1.8 on how this string is constructed
+    # verifier. See RFC4034 section 6 on how this string is constructed
 
     # This method is called by the method that creates as signature
     # and by the method that verifies the signature. It is assumed
@@ -797,7 +864,7 @@ sub _CreateSigData {
     $sigdata= $rdatawithoutsig;
 
 
-    if ( ! $sigzero ){  
+    if ( ! $sigzero ){
 	# Not a SIG0
 	if (@{$rawdata}>1) {
 	    my @canonicaldataarray;
@@ -810,7 +877,7 @@ sub _CreateSigData {
 		$rawdata->[$i]->{"ttl"}=$self->orgttl;
 		# Some error checking is done to. A RRset is defined by 
 		# Same label,class,qtype
-		if ($rawdata->[$i]->name ne $rawdata->[0]->name){
+		if (lc($rawdata->[$i]->name) ne lc($rawdata->[0]->name)){
 		    print "\nError:\n";
 		    for  (my $j=0; $j<@{$rawdata}; $j++){
 			print "\n";
@@ -936,6 +1003,11 @@ sub _checktimeformat {
 }
 
 
+# The previous versions had a typo... *Sigh*
+sub siginceptation {
+    my $self=shift;
+    return $self->siginception(@_);
+}
 
 
 1;
@@ -947,7 +1019,7 @@ Net::DNS::RR::RRSIG - DNS RRSIG resource record
 
 =head1 SYNOPSIS
 
-C<use Net::DNS::RR>;
+C<use Net::DNS::RR;>
 
 =head1 DESCRIPTION
 
@@ -957,7 +1029,8 @@ regular methods in the Net::DNS::RR the Class contains a method to
 sign RRsets using private keys (create). And a class for verifying
 signatures over RRsets (verify).
 
-The RRSIG RR is an implementation of RFC 2535 and RFC 2931.
+The RRSIG RR is an implementation of RFC 4034. 
+See L<Net::DNS::RR::SIG> for an impelementation of SIG0 (RFC 2931).
 
 
 
@@ -966,7 +1039,7 @@ The RRSIG RR is an implementation of RFC 2535 and RFC 2931.
 
 =head2 create
     
-Create a signature over a RR set or over a packet (SIG0).
+Create a signature over a RR set.
 
     my $keypath= 
             "/home/olaf/keys/Kbla.foo.+001+60114.private";
@@ -974,14 +1047,25 @@ Create a signature over a RR set or over a packet (SIG0).
 					$keypath);
     my $sigrr= create Net::DNS::RR::RRSIG(\@datarrset,
 					$keypath,
-					\%arguments);
+					%arguments);
     $sigrr->print;
 
+
+
+    #Alternatively use Net::DNS::SEC::Private 
+
+    my $private=Net::DNS::SEC::Private-new(
+	"/home/olaf/keys/Kbla.foo.+001+60114.private");
+    my $sigrr= create Net::DNS::RR::RRSIG(\@datarrset,
+					  $private);
+
+
+
+
 create is an alternative constructor for a RRSIG RR object.  
 
 The first argument is either reference to an array that contains the
-RRset that needs to be signed or a string containing the data over
-wich a RRSIG0 type of signature needs to be constructed.
+RRset that needs to be signed.
 
 The second argument is a string containing the path to a file
 containing the the private key as generated with dnssec-keygen, a
@@ -991,7 +1075,7 @@ The third argument is an anonymous hash containing the following
 possible arguments:  
 
     ( ttl => 3600,                        # TTL 
-      sigin =>   20010501010101,          # signature inceptation 
+      sigin =>   20010501010101,          # signature inception 
       sigex =>   20010501010101,          # signature expiration
       sigval => 1.5                       # signature validity
       )
@@ -1006,16 +1090,12 @@ then sigex is ignored. The default for sigval is 5 minutes for SIG0s
 and 30 days other types of signatures.
 
 
-Note that for SIG0 signatures the default sigin is calculated at the
-moment the object is created, not at the moment that the packet is put
-on the wire (with ....). So do not leave the object hanging around for
-more than a couple of seconds before sending it.
 
 Notes: 
 
 - Do not change the name of the file generated by dnssec-keygen, the
-create method uses the filename as generated by dnssec-keygen to determine 
-the keyowner, algorithm and the keyid (keytag).
+  create method uses the filename as generated by dnssec-keygen to
+  determine the keyowner, algorithm and the keyid (keytag).
 
 - Only RSA signatures (algorithm 1) and DSA signatures (algorithm 3)
   have been implemented.
@@ -1053,9 +1133,9 @@ Returns the RRs the original TTL of the signature
 
 Returns the expiration date of the signature
 
-=head2 siginceptation
+=head2 siginception
 
-    print "siginceptation =", $rr->siginceptation, "\n"
+    print "siginception =", $rr->siginception, "\n"
 
 Returns the date the signature was incepted.
 
@@ -1071,7 +1151,6 @@ Read "KeyID Bug in bind." below.
     print "signame =", $rr->signame, "\n"
 
 Returns the name of the public KEY RRs  this sig was made with.
-(Note: the name does not contain a trailing dot.)
 
 =head2 sig
 
@@ -1083,6 +1162,8 @@ Returns the base64 representation of the signature.
 =head2 verify and vrfyerrstr
 
     $sigrr->verify($data, $keyrr) || croak $sigrr->vrfyerrstr;
+    $sigrr->verify($data, [$keyrr, $keyrr2, $keyrr3]) || 
+                  croak $sigrr->vrfyerrstr;
 
 
 If $data contains a reference to an array of RR objects then them
@@ -1090,11 +1171,14 @@ method verifies the RRset against the signature contained in the
 $sigrr object itself using the public key in $keyrr.  Because of the
 KeyID bug in bind (see below) a check on keyid is not performed.
 
-
 If $data contains a reference to a Net::DNS::Packet and if $sig->type
 equals zero a a sig0 verification is performed. Note that the
 signature needs to be 'popped' from the packet before verifying.
 
+The second argument can either be a Net::DNS::RR::KEYRR object or a
+reference to an array of such objects. Verification will return
+successful as soon as one of the keys in the array leads to positive
+validation.
 
 Returns 0 on error and sets $sig->vrfyerrstr
 
@@ -1134,8 +1218,8 @@ requests.
 
 =head1 COPYRIGHT
 
-Copyright (c) 2001, 2002  RIPE NCC.  Author Olaf M. Kolkman 
-<net-dns-sec@ripe.net>
+Copyright (c) 2001 - 2005  RIPE NCC.  Author Olaf M. Kolkman 
+<olaf@net-dns.org>
 
 All Rights Reserved
 
@@ -1163,11 +1247,12 @@ This code uses Crypt::OpenSSL which uses the openssl library
 
 =head1 SEE ALSO
 
+L<http://www.net-dns.org/> 
+
 L<perl(1)>, L<Net::DNS>, L<Net::DNS::Resolver>, L<Net::DNS::Packet>,
 L<Net::DNS::Header>, L<Net::DNS::Question>,
-L<Net::DNS::RR>,L<Crypt::OpenSSL::RSA>,L<Crypt::OpenSSL::DSA> RFC 2435 Section
-4, RFC 2931.
+L<Net::DNS::RR>,L<Crypt::OpenSSL::RSA>,
+L<Crypt::OpenSSL::DSA>, L<Net::DNS::SEC::Private>, RFC 4034
 
 =cut
 
-
